# -*- coding: japanese-shift-jis-dos; tab-width: 4; -*-
#
# wdic-fpw - FreePWING script for WDIC 2001
# $Id: wdic-fpw.pl 1.3.0.5 2002/12/13 23:42:37 satomii Exp $
#
# DO NOT CHANGE THE KANJI CODE OF THIS FILE.
#
# Copyright (C) 2000-2002, rei@wdic.org.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
require 5.005;

use strict;
use Getopt::Long;
use FreePWING::FPWUtils::FPWParser;
use File::DosGlob 'glob';
use Jcode;
use Wdic;

require 'wdic-menu.pl';
require 'wdic-body.pl';

use vars qw(
	$VERSION $SJIS_ANY $SJIS_HIRAGANA
	$wdic_dir $opt_charset $opt_bullet $opt_gaiji $opt_cgraphs $opt_plugin
	%cgraphs %plugins
	$fpwtext $fpwheading $fpwword2 $fpwkeyword $fpwmenu $fpwcopyright
	$wdic $newline_on_death
);

$VERSION = '$Revision: 1.3.0.5 $';
$wdic_dir = '.';
$opt_bullet = '';
$opt_charset = '';
$opt_gaiji = 1;
$opt_cgraphs = '';
$opt_plugin = '';
$newline_on_death = 0;

#
# regular expressions for re-formatting.
#
$SJIS_ANY = '(?:[\x81-\x9f\xe0-\xfc][\x40-\x7e\x80-\xfc])';
$SJIS_HIRAGANA = '(?:\x82[\x9f-\xf1])';

my $ASCII_NOSPACE_BOTH = '[\x00-\x20\+\-\*\/\=_~@\#\|]';
my $ASCII_NOSPACE_BEFORE = '[,\.\!\?:;\)\]\>\}]' . "|$ASCII_NOSPACE_BOTH";
my $ASCII_NOSPACE_AFTER = '[\(\[\<\{]' . "|$ASCII_NOSPACE_BOTH";

my $SJIS_X0201 = '[\xa1-\xdf]';
my $SJIS_NOSPACE_BOTH =
	# [{|}~ځہ܁݁ށ߁]
	'(?:\x81[\x7b-\x87\x8b-\x8d\x99\xb8-\xbf\xda-\xe5])|' .
	# [-]
	'(?:\x83[\x9f-\xd6])|' .
	# [@--]
	'(?:\x84[\x40-\xbe])';
my $SJIS_NOSPACE_BEFORE =
	# ABEFGHIJKL^fhjln
	'(?:\x81[\x41\x42\x45-\x4c\x5e\x66\x68\x6a\x6c\x6e' .
	# prtvxz
	'\x70\x72\x74\x76\x78\x7a\x93\x97\xf0-\xf3])' .
	"|$SJIS_NOSPACE_BOTH";
my $SJIS_NOSPACE_AFTER =
	# ABEF^Mefgikmo
	# 'f' is necessary here, it may be placed within a western word.
	'(?:\x81[\x41\x42\x45\x46\x4d\x5e\x65-\x67\x69\x6b\x6d\x6f' .
	# qsuwy
	'\x71\x73\x75\x77\x79\x8f-\x92\x97\x98\x9a-\x9e\x9f-\xa7])' .
	"|$SJIS_NOSPACE_BOTH";

my $NOSPACE_BOTH = "(?:$ASCII_NOSPACE_BOTH|$SJIS_NOSPACE_BOTH|$SJIS_X0201)";
my $NOSPACE_BEFORE = "(?:$ASCII_NOSPACE_BEFORE|$SJIS_NOSPACE_BEFORE|$SJIS_X0201)";
my $NOSPACE_AFTER = "(?:$ASCII_NOSPACE_AFTER|$SJIS_NOSPACE_AFTER|$SJIS_X0201)";

#
# signal hooks.
#
$SIG{__DIE__} = sub {
	my $err = $_[0];

	$err = Jcode::convert($err, $opt_charset) if ($err =~ /[^\x00-\x7e]/);
	$err = "Error: $err";
	$err = "\n$err" if ($newline_on_death);
	die $err;
};

$SIG{__WARN__} = sub {
	my $err = $_[0];

	$err = Jcode::convert($err, $opt_charset) if ($err =~ /[^\x00-\x7e]/);
	$err = "Warning: $err";
	if ($newline_on_death) {
		$err = "\n$err";
		$newline_on_death = 0;
	}
	warn $err;
};

##-------------------------------------------------------------------------
## ENTRY POINT

my($word_count);

&initialize();

$Wdic::CHARSET = $opt_charset;
$wdic = new Wdic;
$word_count = $wdic->init_wdic($wdic_dir);

initialize_fpwparser(
	'text' => \$fpwtext,
	'heading' => \$fpwheading,
	'word2' => \$fpwword2,
	'keyword' => \$fpwkeyword,
	'menu' => \$fpwmenu,
	'copyright' => \$fpwcopyright
);

&make_menu();
&make_copyright();
&make_body($word_count);
&make_plugin_list();

finalize_fpwparser(
	'text' => \$fpwtext,
	'heading' => \$fpwheading,
	'word2' => \$fpwword2,
	'keyword' => \$fpwkeyword,
	'menu' => \$fpwmenu,
	'copyright' => \$fpwcopyright
);
$wdic->uninit_wdic();
exit;

##-------------------------------------------------------------------------
## INITIALIZATION

#
# void initialize(void);
#
sub initialize()
{
	my $opt_usage;

	if (!GetOptions('help|h|?' => \$opt_usage,
			'charset|c:s' => \$opt_charset,
			'bullet|b:s' => \$opt_bullet,
			'gaiji|g:i' => \$opt_gaiji,
			'cgraphs:s' => \$opt_cgraphs,
			'plugin:s' => \$opt_plugin)
			|| 1 < scalar(@ARGV)) {
		&usage(1);
	}
	&usage(0) if ($opt_usage);

	#
	# check the input directory is valid and contains WDIC.MAN. also
	# the current directory must contain the readme.txt file.
	#
	$wdic_dir = shift(@ARGV) if (@ARGV);
	$wdic_dir =~ s/\\/\//g;
	$wdic_dir =~ s/\/$//;
	-d $wdic_dir || die "Invalid input directory: $wdic_dir ($!).\n";

	if (!glob("$wdic_dir/WDIC.MAN") && !(-f "$wdic_dir/WDIC.MAN")) {
		die "File not found: $wdic_dir/WDIC.MAN ($!)\n";
	}
	if (!glob("./readme.txt") && !(-f "./readme.txt")) {
		die "File not found: ./readme.txt ($!)\n";
	}

	if ($opt_plugin) {
		$opt_plugin =~ s/\\/\//g;
		$opt_plugin =~ s/\/$//;
		-d $opt_plugin || die "Invalid plugin directory: $opt_plugin ($!).\n";
	}

	#
	# validate the charset option, or use the environment value if not
	# specified at the command line.
	#
	if (!$opt_charset) {
		if ($ENV{LANG} && $ENV{LANG} =~ /^ja_JP\.([a-zA-Z]+)$/) {
			$opt_charset = $1;
		} else {
			$opt_charset = '';
		}
	}
	$opt_charset =~ tr/A-Z/a-z/;
	if ('ujis' eq $opt_charset || 'eucjp' eq $opt_charset) {
		$opt_charset = 'euc';
	} elsif (!$opt_charset) {
		if ((!$ENV{OS}
				|| $ENV{OS} !~ /^Windows/)
				&& (!$ENV{COMSPEC}
				|| $ENV{COMSPEC} !~ /(COMMAND\.COM)|(CMD\.EXE)$/i)) {
			$opt_charset = 'euc';
		} else {
			$opt_charset = 'sjis';
		}
	} elsif ('sjis' ne $opt_charset
			&& 'euc' ne $opt_charset
			&& 'jis' ne $opt_charset
			&& 'iso_2022_jp' ne $opt_charset) {
		die "Unknown character set: $opt_charset\n";
	}
	print "Detected display charset: $opt_charset\n";

	#
	# ensure that other option values are valid.
	#
	$opt_bullet = '' if (!defined($opt_bullet) || '' eq $opt_bullet);
	$opt_bullet = Jcode::convert($opt_bullet, 'euc');

	#
	# pre-load the cgraphs.txt file is necessary. keys of %cgraphs
	# indicate the color graphic IDs.
	#
	if ($opt_cgraphs) {
		open(CGRAPHS, "<$opt_cgraphs")
			|| die "Unable to open file: $opt_cgraphs\n";
		print "Reading $opt_cgraphs...";
		while (<CGRAPHS>) {
			my @list = split(/[ \t]+/, $_);
			$cgraphs{$list[0]}{name} = '';
		}
		close(CGRAPHS);
		print " done.\n";
	}
}

#
# char *reform_sjis(char *in, int mode)
# reformat SJIS string, inserting white spaces between double- and
# single-byte characters, converting some DBCS into SBCS. but this
# routine may still contain some improper operation.
#
sub reform_sjis($$)
{
	# mode == 0: default, mode == 1: menu, mode == 2: keyword.
	my ($in, $mode) = @_;
	my (@list, $out, $c1, $c2, $prev, @quot);

	@list = unpack('C*', $in);
	$out = '';
	$prev = '';

	while ($c1 = shift(@list)) {
		$c2 = 0;
		$in = substr($in, 1);

		if (0x81 == $c1) {
			$c2 = shift(@list);
			$in = substr($in, 1);
			if (0x40 == $c2) {							# '@'
				$c1 = 0x20; $c2 = 0;
			} elsif (0x48 == $c2) {						# H
				$c1 = 0x3f; $c2 = 0;
			} elsif (0x49 == $c2) {						# I
				$c1 = 0x21; $c2 = 0;
			} elsif (0x5e == $c2) {						# ^
				$c1 = 0x2f; $c2 = 0;
			} elsif (0x69 == $c2 || 0x6a == $c2) {		# ij
				$c1 = $c2 - 0x41; $c2 = 0;
			} elsif (0x66 == $c2) {						# f
				if (2 == $mode
						|| $in =~ /^[0-9][0-9][^0-9]*/
						|| $in =~ /^\x82[\x4f-\x58]\x82[\x4f-\x58]/) {
					$c1 = 0x27; $c2 = 0;
				}
			} elsif (0x93 == $c2) {						# 
				if (2 == $mode || $prev =~ /^[0-9]+$/) {
					$c1 = 0x25; $c2 = 0;
				}
			} elsif ($mode < 2) {
				if (0x46 == $c2) {						# F
					if ("\x97\xe1" eq $prev) {			# 
						$c1 = 0x3a; $c2 = 0x20;
					}
				} elsif (1 == $mode) {
					if (0x41 == $c2) {					# A
						$c1 = 0x2c; $c2 = 0x20;
					} elsif (0x42 == $c2) {				# B
						$c1 = 0x2e; $c2 = 0x20;
					}
				}
			} else {									# keyword mode only.
				if (0x43 == $c2) {						# C
					$c1 = 0x2c; $c2 = 0;
				} elsif (0x44 == $c2) {					# D
					$c1 = 0x2e; $c2 = 0;
				} elsif (0x46 == $c2 || 0x47 == $c2) {	# FG
					$c1 = $c2 - 0x0c; $c2 = 0;
				} elsif (0x62 == $c2) {					# b
					$c1 = 0x7c; $c2 = 0;
				} elsif (0x68 == $c2) {					# h
					$c1 = 0x22; $c2 = 0;
				} elsif (0x6d == $c2) {					# m
					$c1 = 0x5b; $c2 = 0;
				} elsif (0x6e == $c2) {					#  n
					$c1 = 0x5d; $c2 = 0;
				} elsif (0x7b == $c2) {					# {
					$c1 = 0x2b; $c2 = 0;
				} elsif (0x7c == $c2) {					# |
					$c1 = 0x2d; $c2 = 0;
				} elsif (0x96 == $c2) {					# 
					$c1 = 0x2a; $c2 = 0;
				} elsif (0x93 == $c2) {					# 
					$c1 = 0x25; $c2 = 0;
				} elsif (0x94 == $c2) {					# 
					$c1 = 0x23; $c2 = 0;
				}
			}
		} elsif (0x82 == $c1) {
			$c2 = shift(@list);
			$in = substr($in, 1);
			if ((0x4f <= $c2 && $c2 <= 0x58)
					|| (0x60 <= $c2 && $c2 <= 0x79)) {	# [O-X`-y]
				$c1 = $c2 - 0x1f; $c2 = 0;
			} elsif (0x81 <= $c2 && $c2 <= 0x9a) {		# [-]
				$c1 = $c2 - 0x20; $c2 = 0;
			}
		} elsif ((0x83 <= $c1 && $c1 <= 0x9f)
				|| (0xe0 <= $c1 && $c1 <= 0xfc)) {		# other SJIS.
			$c2 = shift(@list);
			$in = substr($in, 1);
		} elsif (0xa5 == $c1)	{						# E in single byte.
			if ($prev !~ /^$SJIS_X0201$/ && $in !~ /^$SJIS_X0201/) {
				$c1 = 0x81;
				$c2 = 0x45;
			}
		}

		$c1 = pack('C', $c1);
		$c1 .= pack('C', $c2) if ($c2);
		if (2 <= $mode) {
			$out .= $c1;
			$prev = $c1;
			next;
		}

		if (' ' eq $c1) {
			next if ($prev =~ / $/ || $in =~ /^,/);
		} elsif ($out =~ /\<[!\?\%]$/ && $in =~ /\>/) {
			# possibly DTD or processing-instruction, nop.
		} elsif ($c1 =~ /^[\'\"]$/) {
			if ("\'" eq $c1
					&& ($in =~ /^[0-9][0-9][^0-9]*/
					|| $in =~ /^\x82[\x4f-\x58]\x82[\x4f-\x58]/)) {
				# maybe 2-digit year.
				if ('' ne $prev
						&& $prev !~ /^(?:$NOSPACE_AFTER|[\'\"])$/) {
					$out .= ' ';
				}
			} elsif ("\'" eq $c1 && $prev =~ /^[a-zA-Z]$/) {
				# maybe within some western word, nop.
			} elsif (@quot && $c1 eq $quot[$#quot]) {
				# quotation closed.
				pop @quot;
				if ($mode < 2
						&& '' ne $in
						&& $in !~ /^(?:$NOSPACE_BEFORE|[\'\"])/) {
					$c1 .= ' ';
				}
			} else {
				# quotation opened.
				push(@quot, $c1);
				if ('' ne $prev
						&& $prev !~ /^(?:$NOSPACE_AFTER|[\'\"])$/) {
					$out .= ' ';
				}
			}
		} elsif ('' eq $prev || $prev =~ / $/) {
			# no need to adjust whitespace.
		} elsif ('.' eq $c1
				&& $prev =~ /^(?:$SJIS_HIRAGANA|(?:\x8e\x71))$/
				&& $in =~ /^[A-Za-z0-9_~]+(?:[ ,\)]|$SJIS_HIRAGANA)/) {
			# possibly filename extension.
			$out .= ' ';
		} elsif (',' eq $c1) {
			if ($prev =~ /^[0-9]$/
					&& $in =~ /^[0-9]{3}/
					&& $out !~ /[0-9]+, [0-9]+$/) {
				# digit delimitor, nop.
			} elsif ($in !~ /^(?:$NOSPACE_BEFORE|[\'\"])/) {
				$c1 .= ' ';
			}
		} elsif ('?' eq $c1) {
			if ($out =~ /[0-9]\)? `$/ && $in =~ /^\./) {
				# living years.
				$out .= ' ';
			}
		} elsif (':' eq $c1) {
			if ($prev =~ /$SJIS_ANY/ ||
					($out =~ /\([a-zA-Z]+$/ &&
					($in =~ /^[a-zA-Z]{2}[^ ]* [a-zA-Z]/
					|| $in =~ /^$SJIS_ANY/))) {
				$c1 .= ' ';
			}
		} elsif ('-' eq $c1) {
			if ('-' eq $prev && $in =~ /^$SJIS_ANY/) {
				$c1 .= ' ';
			} elsif ($prev =~ /$SJIS_ANY/
					&& ($in =~ /^[0-9\.,]+[a-zA-Z]*(?:$SJIS_ANY| )/
					|| $in =~ /^\-+($SJIS_ANY| )/)) {
				$out .= ' ';
			}
		} elsif ('+' eq $c1) {
			if ($prev =~ /$SJIS_ANY/
					&& $in =~ /^[0-9\.,]+[a-zA-Z]*(?:$SJIS_ANY| )/) {
				$out .= ' ';
			} elsif ($in =~ /^\+/ && '' ne $prev) {
				if ($prev =~ /$SJIS_ANY/
						&& $prev !~ /(?:$SJIS_NOSPACE_AFTER|[\'\"])/) {
					$out .= ' ';
				}
			} elsif ('+' eq $prev && $in =~ /^(?:$SJIS_ANY|\()/) {
				$c1 .= ' ';
			}
		} elsif ('(' eq $c1) {
			if ($out =~ / main$/) {
				# main(), nop.
			} elsif ($out =~ / [a-zA-Z]$/ && $in =~ /^[a-zA-Z]\)/) {
				# like f(x) or r(t), nop.
			} elsif ('' eq $prev
					|| $prev =~ /^(?:$NOSPACE_AFTER|[\'\"])$/
					|| $out =~ /sin$/
					|| $out =~ /cos$/
					|| $out =~ /log[0-9]?$/
					|| $out =~ /[^a-zA-Z]Z$/
					|| $in =~ /^x\)/
					|| $in =~ /^TM\) /) {
				# also nop.
			} elsif ($in =~ /^R\)/
					&& ($out =~ /Intel$/i
					|| $out =~ /Pentium$/i
					|| $out =~ /Ce$/)) {
				# (R), nop.
			} else {
				$out .= ' ';
			}
		} elsif (')' eq $c1) {
			if ($in !~ /^(?:$NOSPACE_BEFORE|[\'\"])/
					&& $out !~ /log\([^\)]+$/
					&& $in !~ /^\^[0-9]/) {
				$c1 .= ' ';
			}
		} elsif ('<' eq $c1) {
			if ($out !~ /\>/ && $in !~ /\>/) {
				if ($prev !~ /^[\x20-\x7e]$/ || $in !~ /^[\x20-\x7e]/) {
					if ($prev !~ /^$NOSPACE_AFTER$/) {
						$out .= ' ';
					}
					if ($in !~ /^$NOSPACE_BEFORE/) {
						$c1 .= ' ';
					}
				}
			} elsif ($in =~ /^\/([A-Za-z][A-Za-z0-9]*)\>/) {
				my $name = $1;
				if ($out =~ /\<$name(?: [^\<\>]*)?\>/) {
					# possibly SGML element, nop.
				} elsif ($prev !~ /^$NOSPACE_AFTER$/
						&& $prev =~ /$SJIS_ANY$/) {
					$out .= ' ';
				}
			} elsif ($prev !~ /^[\x20-\x7e]$/
					&& $prev !~ /^$NOSPACE_AFTER$/) {
				$out .= ' ';
			}
		} elsif ('>' eq $c1) {
			if ($out =~ /\<([A-Za-z][A-Za-z0-9]*)(?: [^\<\>]*)?$/) {
				my $name = $1;
				if ($in =~ /<\/$name>/) {
					# SGML element, nop.
				} elsif ($in !~ /^[\x20-\x7e]/ && $in !~ /^$NOSPACE_BEFORE/) {
					$c1 .= ' ';
				}
			} elsif ($prev !~ /^[\x20-\x7e]$/
					&& $in =~ /^>[0-9]+ *$SJIS_ANY/) {
				# 2ch link.
				$out .= ' ';
			} elsif ($in =~ /^$SJIS_ANY/ && $in !~ /^$NOSPACE_BEFORE/) {
				$c1 .= ' ';
			}
		} elsif ('@' eq $c1 || '#' eq $c1) {
			if (($prev =~ /(?:$SJIS_HIRAGANA|\x81\x60||\x8c[\xb3\xbb])$/
					&& $in =~ /^[A-Za-z0-9_]*(?:[ ,\)]|$SJIS_HIRAGANA|\x81[\x41-\x44\x48\x49\x60])/)
					|| ($prev =~ /$SJIS_ANY$/
					&& $in =~ /^[0-9]+\b/)) {
				$out .= ' ';
				if ($in =~ /^[^\x20-\x7e]/ && $in !~ /^$SJIS_NOSPACE_BEFORE/) {
					$c1 .= ' ';
				}
			}
		} elsif ('' eq $prev
				 || $prev =~ /^(?:$NOSPACE_AFTER|[\'\"])$/
				 || $c1 =~ /^$NOSPACE_BEFORE$/) {
			# nop.
		} elsif ($prev =~ /^[\x21-\x7e]$/) {
			# previous character is ASCII.
			if ($c1 =~ /^(?:$SJIS_ANY|[\(\[])$/
					&& $out !~ /@[^\"]+\.$/) {
				$out .= ' ';
			}
		} else {
			# SJIS double bytes.
			if ($c1 =~ /^[\x21-\x7e]$/) {
				$out .= ' ';
			}
		}

		$out .= $c1;
		$prev = $c1;
	}

	return $out;
}

#
# bool write_sjis($obj, char *in)
# writes SJIS text using the specified class object.
#
sub write_sjis($$)
{
	my ($fpwobj, $in) = @_;
	my ($newline, $out, $c);

	$newline = ($in =~ s/([^\r\n]*)[\r\n]+$/$1/);

	if ($opt_gaiji < 2) {
		# do not use user characters.
		$out = Jcode::convert($in, 'euc', 'sjis');
		$out =~ s/\\/\xa1\xef/g;	#  in EUC.
		$fpwobj->add_text($out)
			|| die &death_cry("Failed to write text: $out", $fpwobj);
		if ('' eq $in || $newline) {
			$fpwobj->add_newline()
				|| die &death_cry("Unable to add newline", $fpwobj);
		}
		return 1;
	}

	$out = '';

	while ('' ne $in) {
		$c = substr($in, 0, 1);
		$in = substr($in, 1);

		if ($c =~ /[\x81-\x9f\xe0-\xfc]/) {
			# SJIS double byte characters.
			$out .= $c;
			$out .= substr($in, 0, 1);
			$in = substr($in, 1);

		} elsif ($c =~ /[\x5c\xa1-\xdf]/) {
			# '\' or JIS X 0201.
			my $gaiji = sprintf("%02x", unpack('C', $c));
			if ('' ne $out) {
				$fpwobj->add_text(Jcode::convert($out, 'euc', 'sjis'))
					|| die &death_cry("Failed to write text: $out", $fpwobj);
				$out = '';
			}
			$fpwobj->add_half_user_character($gaiji)
				|| die &death_cry("Unable to embed user char: $gaiji", $fpwobj);
		} else {
			# ASCII characters.
			$out .= $c;
		}
	}

	if ('' ne $out) {
		$fpwobj->add_text(Jcode::convert($out, 'euc', 'sjis'))
			|| die &death_cry("Failed to write text: $out", $fpwobj);
	}
	if ('' eq $_[1] || $newline) {
		$fpwobj->add_newline()
			|| die &death_cry("Unable to add newline", $fpwobj);
	}
	1;
}

#
# bool write_newline($fpwobj, int count)
# writes newline using the specified object.
#
sub write_newline($;$)
{
	my ($fpwobj, $count) = @_;
	my $i;

	$count = 1 unless defined($count);
	for ($i = 0; $i < $count; $i++) {
		$fpwobj->add_newline()
			|| die &death_cry("Unable to add newline", $fpwobj);
	}
	$i;
}

#
# bool write_link($fpwobj, char *label, char *tag)
# adds a cross-reference to another entry using the specified class object.
# the label is assumed to be SJIS string.
#
sub write_link($$$)
{
	my ($fpwobj, $label, $tag) = @_;
	my $newline = ($label =~ s/([^\r\n])[\r\n]+$/$1/);
	if (!$fpwobj->add_reference_start()
			|| !$fpwobj->add_text(Jcode::convert($label, 'euc', 'sjis'))
			|| !$fpwobj->add_reference_end($tag)) {
		die &death_cry("Unable to add reference: $label", $fpwobj);
	}
	&write_newline($fpwobj, 1) if ($newline);
	return 1;
}

#
# char **write_word_link($fpwobj,
#      char *bullet, char *word, char *group, char *file, int mode)
# searches another word in the current source files and links to it if
# one is found. returns the target group and file (dictionary) if found.
#
sub write_word_link($$$$$)
{
	my ($fpwobj, $bullet, $word, $group, $file) = @_;
	my (@list, $info, $newline);

	$newline = ($word =~ s/([^\r\n])[\r\n]+$/$1/);
	$info = '';

	# @list will be: (index, group, file, offset)
	@list = $wdic->locate_word($word, $group, $file);

	if (4 == scalar(@list)) {
		$info = sprintf("%s%08x", $list[2], $list[3]);
		&write_sjis($fpwobj, $bullet) if ('' ne $bullet);
		&write_link($fpwobj, $word, $info);
		@list = ($list[1], $list[2]);
	} else {
		&write_sjis($fpwobj, "$bullet$word");
	}

	&write_newline($fpwobj, 1) if ($newline);
	return @list;
}

##-------------------------------------------------------------------------
## MISC FUNCTIONS

sub death_cry($$)
{
	my ($str, $fpwobj) = @_;

	if ($fpwobj) {
		my $err = $fpwobj->error_message();
		$str .= " ($err)" if ($err);
	}
	return "$str,";
}

sub usage($)
{
	eval { require Pod::Usage };
	if ($@) {
		print<<__EOF__;
Usage: perl $0 [-c CHARSET] [-b BULLET] [-g LEVEL]
	    [--cgraphs=FILE] [--plugin=DIR] [INPUT_DIR]
If inputdir is omitted, the current directory (.) is used.
For more information, use pod2man(1) or any similar utility.

__EOF__
	} else {
		Pod::Usage::pod2usage(-verbose => 1);
	}
	exit($_[0]);
}

__END__

##-------------------------------------------------------------------------
## DOCUMENTATION

=pod

=head1 NAME

wdic-fpw.pl - a FreePWING script for WDIC

=head1 SYNOPSIS

perl wdic-fpw.pl [-c CHARSET] [-b BULLET] [-g GAIJI] [--cgraphs=FILE]
[--plugin=DIR} [INPUT_DIR]

=head1 DESCRIPTION

This script generates a JIS X 4081 book from 'WDIC' (possibly
"WIRED & WIRELESS DICTIONARY") using FreePWING. The generated book can be
viewed with many EPWING viewers.

=head1 OPTIONS

=over 8

=item B<-h> | B<-?> | B<--help>

Displays this usage.

=item B<-c> I<CHARSET> | B<--charset>=I<CHARSET>

Specifies the display charset, which is used when printing progress. The
default is I<sjis> for Windows and I<euc> for others. This option would never
affect the book-producing routines.

=item B<-b> I<BULLET> | B<--bullet>=I<BULLET>

Specifies a character or string to be used as a 'bullet' for each description.
The default is a full-width white diamond.

=item B<-g> I<LEVEL> | B<--gaiji>=I<LEVEL>

Use half-shape user-defined characters (gaiji). Valid levels are:
0) Do not use any user characters;
1) Use user characters for some phonetic symbols;
2) Use user characters for some phonetic symbols, signle-byte yen marks ('\')
and so-colled 'hankaku katakana's.

=item B<--cgraphs>=I<FILE>

Specifies the path to the cgraph.txt file.
If this option is used, WDIC plugin images found in FILE is embedded into the
resulting book.
FILE must be the one generated by Makefile that comes with this script.

=item B<--plugin>=I<DIR>

Specifies the directory where plugin files are located.
Some text-based plugin files from this directory may be embedded to the book.

=back

=head1 B<ENVIRONMENT>

=over 8

=item B<LANG>

Referred when B<--charset> option is not given.

=item B<OS> or B<COMSPEC> or B<ComSpec>

Referred when B<--charset> option is not given and B<LANG> is not defined, to
distinguish the Windows systems.

=item B<USER> or B<USERNAME>

Used to record the build information into the copyright page. If these
variables were not defined, the builder's name would default to
'(unknown user)'.

=back

=head1 B<FILES>

=item B<FreePWING>

A set of Perl modules to generate a JIS X 4081 book. Visit
E<lt>http://www.sra.co.jp/people/m-kasahr/freepwing/E<gt> for more information.

=item B<Jcode.pm>

A Perl module for kanji-code conversion. Visit
E<lt>http://openlab.ring.gr.jp/Jcode/E<gt> for more information.

=item B<Wdic.pm>

A Perl module to read entries from WDIC files.

=item B<wdic-menu.pl>

A Perl script used from wdic-fpw main script, to construct the menu and
copyright entries.

=item B<wdic-body.pl>

A Perl script used from wdic-fpw main script, to write all word entries into
the book.

And, of course, the original WDIC files are also required to run wdic-fpw.

=head1 VERSION

$Id: wdic-fpw.pl 1.3.0.5 2002/12/13 23:42:37 satomii Exp $

=head1 AUTHOR

rei@wdic.org

=head1 SEE ALSO

=item B<WIRED & WIRELESS DICTIONARY>

http://www.wdic.org/

=cut
